@@ -37,7 +37,7 @@ def dec_tuple(o):
         if o[0] == 'list':
             return list(dec_tuple(p) for p in o[1:])
         else:
-            assert(o[0]) == 'tuple'
+            assert(o[0] == 'tuple')
             return tuple(dec_tuple(p) for p in o[1:])
     else:
         return o
@@ -126,7 +126,7 @@ class GraphMapping(MutableMapping):
         seen = set()
         for (branch, rev) in self.gorm._active_branches():
             data = self.gorm.cursor.execute(
-                "SELECT graph_val.key "
+                "SELECT graph_val.key, graph_val.value "
                 "FROM graph_val JOIN ("
                 "SELECT graph, key, branch, MAX(rev) AS rev FROM graph_val "
                 "WHERE graph=? "
@@ -136,8 +136,7 @@ class GraphMapping(MutableMapping):
                 "ON graph_val.graph=hirev.graph "
                 "AND graph_val.key=hirev.key "
                 "AND graph_val.branch=hirev.branch "
-                "AND graph_val.rev=hirev.rev "
-                "WHERE graph_val.key IS NOT NULL;",
+                "AND graph_val.rev=hirev.rev;",
                 (
                     self.graph._name,
                     branch,
@@ -146,11 +145,10 @@ class GraphMapping(MutableMapping):
             ).fetchall()
             if len(data) == 0:
                 continue
-            for row in data:
-                key = json_load(row[0])
-                if key not in seen:
-                    yield key
-                seen.add(key)
+            for (k, v) in data:
+                if k not in seen and v is not None:
+                    yield json_load(k)
+                seen.add(k)
 
     def __contains__(self, k):
         """Do I have a value for this key right now?"""
@@ -181,7 +179,7 @@ class GraphMapping(MutableMapping):
             return data[0][0] is not None
 
     def __len__(self):
-        """Number of non-'unset' keys"""
+        """Number of set keys"""
         n = 0
         for k in iter(self):
             n += 1
@@ -218,6 +216,8 @@ class GraphMapping(MutableMapping):
                 continue
             elif len(results) > 1:
                 raise ValueError("Silly data in graph_val table")
+            elif results[0][0] is None:
+                raise KeyError("Key is not set now")
             else:
                 return json_load(results[0][0])
         raise KeyError("key is not set, ever")
@@ -334,6 +334,15 @@ class GraphMapping(MutableMapping):
         rev = self.gorm.rev
         return self.window(self.gorm.branch, rev - revs, rev)
 
+    def update(self, other):
+        """Version of ``update`` that doesn't clobber the database so much"""
+        for (k, v) in other.items():
+            if (
+                    k not in self or
+                    self[k] != v
+            ):
+                self[k] = v
+
 
 class GraphNodeMapping(GraphMapping):
     """Mapping for nodes in a graph"""
@@ -2138,3 +2147,35 @@ class MultiDiGraph(networkx.MultiDiGraph, GormGraph):
             (u, v) = e[:2]
             if u in self.succ and v in self.succ[u]:
                 del self.succ[u][v]
+
+    def add_edge(self, u, v, key=None, attr_dict=None, **attr):
+        """Version of add_edge that only writes to the database once."""
+        if attr_dict is None:
+            attr_dict = attr
+        else:
+            try:
+                attr_dict.update(attr)
+            except AttributeError:
+                raise NetworkXError(
+                    "The attr_dict argument must be a dictionary."
+                )
+        if u not in self.node:
+            self.node[u] = {}
+        if v not in self.node:
+            self.node[v] = {}
+        if v in self.succ[u]:
+            keydict = self.adj[u][v]
+            if key is None:
+                key = len(keydict)
+                while key in keydict:
+                    key += 1
+            datadict = keydict.get(key, {})
+            datadict.update(attr_dict)
+            keydict[key] = datadict
+        else:
+            if key is None:
+                key = 0
+            datadict = {}
+            datadict.update(attr_dict)
+            keydict = {key: datadict}
+            self.succ[u][v] = keydict
